from conda_store_server.server.views.api import app_api  # noqa
from conda_store_server.server.views.registry import app_registry  # noqa
from conda_store_server.server.views.ui import app_ui  # noqa
from conda_store_server.server.views.metrics import app_metrics  # noqa
